/*
 * key-draw.c -- draw the keyboard and highlight the keys
 *
 * Copyright (c) 2000 Alessandro Rubini (rubini@gnu.org)
 * Sponsored by Linuxcare Italia Spa.
 *
 * Extensive modifications by Jorj Bauer <jorj@seas.upenn.edu>,
 * sponsored by DejaVu Software, Inc. (http://www.dejavusoftware.com).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <linux/types.h>
#include <sys/ioctl.h>
#include <linux/fb.h>

#include "key-engine.h"

static int fbfd = -1;
static void *fbpointer;
static void *offscreenkeymap;
static int offscreenkeymapsize;
extern struct keydef *kptr;
extern int shiftstate;

extern unsigned char fontdata_8x8[2048];

int kdraw_init(char *fbfile, fbinfo *devinfo)
{
    char *mapaddr;
    struct fb_fix_screeninfo fb_fix;
    struct fb_var_screeninfo fb_var;
    struct keydef *kp;

    if (!fbfile)
	return 0;

    /*
     * should: open the fb file, find info, reserve a region (in the
     * future) draw the keyboard
     */

    fbfd = open(fbfile, O_RDWR);
    if (fbfd<0) {
	fprintf(stderr, "%s: %s\n", fbfile, strerror(errno));
	exit(1);
    }
    /* Get information about the frame buffer. -- jorj */
    if (ioctl(fbfd, FBIOGET_FSCREENINFO, &fb_fix) == -1 ||
	ioctl(fbfd, FBIOGET_VSCREENINFO, &fb_var) == -1) {
      fprintf(stderr, "%s: error reading screen info: %s\n", KE_FBDEV,
	      strerror(errno));
      exit(1);
    }

    devinfo->width = fb_var.xres;
    devinfo->height = fb_var.yres;
    devinfo->depth = fb_var.bits_per_pixel;
    devinfo->linelen = fb_fix.line_length;

    devinfo->size = devinfo->linelen * devinfo->height;
    devinfo->size = (devinfo->size + getpagesize() - 1) 
      / getpagesize() * getpagesize();

    /* Update the keymap file with new dynamic widths. -- jorj */
    if (devinfo->width != 240) {
      printf("Updating keymap width at ratio %f...\n", (float)devinfo->width/240.0);
      for (kp = kptr; kp->special >= 0; kp++) {
	kp->x[0] *= ((float)devinfo->width / 240.0);
	kp->x[1] *= ((float)devinfo->width / 240.0);
	if (kp->y[2]) { /* also do 2-part keys, like 'return' */
	  kp->x[2] *= ((float)devinfo->width / 240.0);
	  kp->x[3] *= ((float)devinfo->width / 240.0);
	}
      }
    }

    offscreenkeymapsize = 97 * devinfo->linelen; /* HACK -- jorj */
    devinfo->offscreenkeymap = offscreenkeymap = calloc(offscreenkeymapsize, 
							sizeof(char));
    if (!offscreenkeymap) {
      fprintf(stderr, "%s: unable to malloc for offscreen keyboard: %s\n",
	      fbfile, strerror(errno));
      exit(1);
    }

    printf("%s info: %d x %d x %d, line is %d bytes\n", 
	   KE_FBDEV, devinfo->width, devinfo->height,
	   devinfo->depth, devinfo->linelen);

    devinfo->addr = mapaddr = mmap(0, devinfo->size, 
				   PROT_READ | PROT_WRITE, MAP_SHARED, 
				   fbfd, 0);
    if (mapaddr == (char *)-1) {
	fprintf(stderr, "mmap(%s): %s\n", fbfile, strerror(errno));
	exit(1);
    }
    fbpointer = (__u16 *)mapaddr;

    draw_keyboard(devinfo);

    memcpy(fbpointer, offscreenkeymap, offscreenkeymapsize);
    return 0;
}

void kdraw_redraw(void)
{
    memcpy(fbpointer, offscreenkeymap, offscreenkeymapsize);
}

/* there are two or four coordinates */
int kdraw_xor(unsigned char *buf, int *xp, int *yp, fbinfo *devinfo)
{
    int i, y;

    if (fbfd<0)
	return 0;
    for(i=0; i<2; i++) { /* two rectangles */
      for (y=yp[0]; y<=yp[1]; y++)
	switch (devinfo->depth) {
	case 1:
	  linear1_drawhline(buf, devinfo, xp[0], xp[1], y, 1, MODE_XOR);
	  break;
	case 2:
	  linear2_drawhline(buf, devinfo, xp[0], xp[1], y, 3, MODE_XOR);
	  break;
	case 4:
	  linear4_drawhline(buf, devinfo, xp[0], xp[1], y, 0xF, MODE_XOR);
	  break;
	case 8:
	  linear8_drawhline(buf, devinfo, xp[0], xp[1], y, 0xFF, MODE_XOR);
	  break;
	case 16:
	  linear16_drawhline(buf, devinfo, xp[0], xp[1], y, 0xFFFF, MODE_XOR);
	  break;
	case 24:
	  linear24_drawhline(buf, devinfo, xp[0], xp[1], y, 0xFFFFFF, MODE_XOR);
	  break;
	case 32:
	  linear32_drawhline(buf, devinfo, xp[0], xp[1], y, 0xFFFFFFFF, MODE_XOR);
	  break;
	default:
	  break;
	}
      yp += 2; xp += 2;
      if (!i && !yp[0]) break; /* no second rectangle */
    }
    return 0;
}

int key_draw_dot(char *offscreenkeymap, fbinfo *devinfo,
		 int x, int y, int on)
{
  switch (devinfo->depth) {
  case 1:
    linear1_drawpixel(offscreenkeymap, devinfo, x, y, on?1:0);
    break;
  case 2:
    linear2_drawpixel(offscreenkeymap, devinfo, x, y, on?3:0);
    break;
  case 4:
    linear4_drawpixel(offscreenkeymap, devinfo, x, y, on?10:0);
    break;
  case 8:
    linear8_drawpixel(offscreenkeymap, devinfo, x, y, on?200:0);
    break;
  case 16:
    linear16_drawpixel(offscreenkeymap, devinfo, x, y, on?0xF000:0);
    break;
  case 24:
    linear24_drawpixel(offscreenkeymap, devinfo, x, y, on?0xF0F0F0:0);
    break;
  case 32:
    linear32_drawpixel(offscreenkeymap, devinfo, x, y, on?0xF0000000:0);
    break;
  default:
    fprintf(stderr, "Unhandled drawing type for %d bpp depth\n", devinfo->depth);
  }
  return 0;
}

int key_draw_rect(char *offscreenkeymap, fbinfo *devinfo, 
		  int xi, int yi, int xe, int ye, int mode)
{
  switch (devinfo->depth) {
  case 1:
    /* draw horizontal lines */
    linear1_drawhline(offscreenkeymap, devinfo, xi, xe, yi, 1, mode);
    linear1_drawhline(offscreenkeymap, devinfo, xi, xe, ye, 1, mode);
    /* draw vertical lines */
    linear1_drawvline(offscreenkeymap, devinfo, yi, ye, xi, 1);
    linear1_drawvline(offscreenkeymap, devinfo, yi, ye, xe, 1);
    break;
  case 2:
    /* draw horizontal lines */
    linear2_drawhline(offscreenkeymap, devinfo, xi, xe, yi, 3, mode);
    linear2_drawhline(offscreenkeymap, devinfo, xi, xe, ye, 3, mode);
    /* draw vertical lines */
    linear2_drawvline(offscreenkeymap, devinfo, yi, ye, xi, 3);
    linear2_drawvline(offscreenkeymap, devinfo, yi, ye, xe, 3);
    break;
  case 4:
    /* draw horizontal lines */
    linear4_drawhline(offscreenkeymap, devinfo, xi, xe, yi, 10, mode);
    linear4_drawhline(offscreenkeymap, devinfo, xi, xe, ye, 10, mode);
    /* draw vertical lines */
    linear4_drawvline(offscreenkeymap, devinfo, yi, ye, xi, 10);
    linear4_drawvline(offscreenkeymap, devinfo, yi, ye, xe, 10);
    break;
  case 8:
    /* draw horizontal lines */
    linear8_drawhline(offscreenkeymap, devinfo, xi, xe, yi, 200, mode);
    linear8_drawhline(offscreenkeymap, devinfo, xi, xe, ye, 200, mode);
    /* draw vertical lines */
    linear8_drawvline(offscreenkeymap, devinfo, yi, ye, xi, 200);
    linear8_drawvline(offscreenkeymap, devinfo, yi, ye, xe, 200);
    break;
  case 16:
    /* draw horizontal lines */
    linear16_drawhline(offscreenkeymap, devinfo, xi, xe, yi, 0xF000, mode);
    linear16_drawhline(offscreenkeymap, devinfo, xi, xe, ye, 0xF000, mode);
    /* draw vertical lines */
    linear16_drawvline(offscreenkeymap, devinfo, yi, ye, xi, 0xF000);
    linear16_drawvline(offscreenkeymap, devinfo, yi, ye, xe, 0xF000);
    break;
  case 24:
    /* draw horizontal lines */
    linear24_drawhline(offscreenkeymap, devinfo, xi, xe, yi, 0xF0F0F0, mode);
    linear24_drawhline(offscreenkeymap, devinfo, xi, xe, ye, 0xF0F0F0, mode);
    /* draw vertical lines */
    linear24_drawvline(offscreenkeymap, devinfo, yi, ye, xi, 0xF0F0F0);
    linear24_drawvline(offscreenkeymap, devinfo, yi, ye, xe, 0xF0F0F0);
    break;
  case 32:
    /* draw horizontal lines */
    linear32_drawhline(offscreenkeymap, devinfo, xi, xe, yi, 0xF0000000, mode);
    linear32_drawhline(offscreenkeymap, devinfo, xi, xe, ye, 0xF0000000, mode);
    /* draw vertical lines */
    linear32_drawvline(offscreenkeymap, devinfo, yi, ye, xi, 0xF0000000);
    linear32_drawvline(offscreenkeymap, devinfo, yi, ye, xe, 0xF0000000);
    break;
  default:
    fprintf(stderr, "Unhandled drawing type for %d bpp depth\n", devinfo->depth);
  }

  return 0;
}

int key_draw_string(char *offscreenkeymap, fbinfo *devinfo,
		    char *s,
		    int xi, int yi, int xe, int ye)
{
  int pos, x;

  x = xi + ((xe-xi)>>1);
  x -= 4*strlen(s);

  for (pos=0;pos<strlen(s);pos++) {
    key_draw_char(offscreenkeymap, devinfo, s[pos], 
		  x+8*pos, yi, x+8*(pos+1), ye);
  }
  return 0;
}


int key_draw_char(char *offscreenkeymap, fbinfo *devinfo, 
		  int c,
		  int xi, int yi, int xe, int ye)
{
  /* the 'fontdata_8x8' array assumes 1 byte per pixel. Since that's
   * not necessarily the case here, we'll draw each dot separately.
   * We're not overly worried about speed anyway, since this is a
   * one-time operation (creating the keyboard bitmap, that is).
   * -- jorj
   */
  
  unsigned char line_to_draw;
  int x, y, centerx, centery;

  centerx = (xi+xe)/2;
  centery = (yi+ye)/2;
  centerx -= 4; /* the chars are 8 pixels wide. */
  centery -= 3; /* center them. */

  for (y=0;y<8;y++) {
    line_to_draw = fontdata_8x8[8*c + y];
    for (x=0;x<8;x++) {
      if (line_to_draw & (1<<(7-x)))
	key_draw_dot(offscreenkeymap, devinfo, x+centerx, y+centery, 1);
      else
	key_draw_dot(offscreenkeymap, devinfo, x+centerx, y+centery, 0);
    }
  }

  return 0;
}

void draw_keyboard(fbinfo *devinfo)
{
  struct keydef *kp;
  int *xp, *yp;
  int i;

  /* Draw a keyboard map based on the keys we've read in from the config
     file. Draw into offscreenkeymap, which we'll then blit onscreen. */

  memset(offscreenkeymap, 0, offscreenkeymapsize);
  
  for (kp = kptr; kp->special >= 0; kp++) {
    xp = kp->x;
    yp = kp->y;
    for (i=0;i<2;i++) {
      key_draw_rect(offscreenkeymap, devinfo, 
		    xp[0], yp[0], xp[1], yp[1], MODE_COPY);
      
      if (kp->codes[0]) {
	if (((shiftstate & SPECIAL_SHIFT) || (shiftstate & SPECIAL_CAPS)) && 
	    kp->codes[1]) {
	  key_draw_char(offscreenkeymap, devinfo, kp->codes[1],
		        xp[0], yp[0], xp[1], yp[1]);
	} else {
	  key_draw_char(offscreenkeymap, devinfo, kp->codes[0],
			xp[0], yp[0], xp[1], yp[1]);
	}
      } else if (kp->special == SPECIAL_CONTROL) {
	int w;
	
	w = (xp[1] - xp[0]) >> 1;
	key_draw_string(offscreenkeymap, devinfo, "CTL",
			xp[0], yp[0], xp[1], yp[1]);
      }
      else if (kp->special && kp->special <= SPECIAL_CAPS)
	key_draw_char(offscreenkeymap, devinfo, kp->special-1,
		      xp[0], yp[0], xp[1], yp[1]);


      xp+=2; yp+=2;
      if (i==0 && !yp[0]) break; // second rectangle doesn't exist.
    }
    
  }
}



